# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

'''
caldavctl uses a simple format to create new objects. this format is parsed and
converted into iCalendar objects.
'''

import pytz

from datetime import datetime
from abc import ABC, abstractmethod


STRING = 0
LIST = 1
DATE = 2
INTEGER = 3
TZ = 4

# States:
KEY = 0
VALUE = 1
COMMENT = 2


class ValueParser(ABC):
    def __init__(self, value):
        self.value = value
        self.tokens = ()

    @abstractmethod
    def tokenize(self):
        ...

    @abstractmethod
    def validate(self):
        ...

    def parse(self):
        self.tokenize()
        self.validate()
        return self.tokens


class ParserError(Exception):
    pass


class ObjectParser:
    def __init__(self, object, lexicon):
        self.object = object
        self.lexicon = lexicon

    def tokenize(self):
        data = self.object
        tokens = {}
        pos = 0
        lenght = len(data)
        state = KEY
        while pos < lenght:
            if data[pos] in ' \t':  # Ignore spaces
                pos += 1
                continue
            if data[pos] == '#':  # Ignore comments
                pos = data.find('\n', pos) + 1
                if pos == 0:  # End of file
                    pos = lenght
                continue

            if state == KEY:  # Find the key
                if data[pos] == '\n':
                    pos += 1
                    continue
                end = data.find(':', pos)
                if end == -1:
                    raise ParserError('Could not find key/value pair.')
                key = data[pos:end].strip().lower()
                pos = end + 1
                state = VALUE
                continue

            if state == VALUE:  # Find the value
                if pos == lenght or data[pos] == '\n':  # End of file or empty value
                    tokens[key] = ''
                    pos += 1
                    state = KEY
                    continue
                if data[pos:pos + 2] == '[[':  # Multi line string
                    end = data.find(']]', pos)
                    value = data[pos + 2:end].strip()
                    tokens[key] = value
                    pos = end + 2
                    state = KEY
                    continue
                else:
                    # Single line string
                    end = data.find('\n', pos)
                    if end == -1:  # End of file
                        end = lenght
                    comment = data.find('#', pos)
                    if comment < end and comment != -1:
                        value = data[pos:comment].strip()
                    else:
                        value = data[pos:end].strip()
                    tokens[key] = value
                    pos = end + 1
                    state = KEY
                    continue
            pos += 1
        return tokens

    def parse(self, tokens):
        result = {}
        for key, value in tokens.items():
            if key not in self.lexicon:
                raise ParserError(f'Unknown key: "{key}".')

            optional = self.lexicon[key][1]
            value_type = self.lexicon[key][0]

            if not value and not optional:
                raise ParserError(f'Key "{key}" is not optional')
            elif not value and optional:
                continue

            if value_type == STRING:
                result[key] = value
            elif value_type == LIST:
                result[key] = [v.strip() for v in value.split(',')]
            elif value_type == DATE:
                try:
                    result[key] = datetime.fromisoformat(value)
                except ValueError:
                    raise ParserError(f'Invalid date format for: "{value}"')
            elif value_type == INTEGER:
                try:
                    result[key] = int(value)
                except ValueError:
                    raise ParserError(f'Invalid integer: "{value}"')
            elif value_type == TZ:
                try:
                    result[key] = pytz.timezone(value)
                except pytz.exceptions.UnknownTimeZoneError:
                    raise ParserError(f'Invalid timezone: "{value}"')
            elif issubclass(value_type, ValueParser):
                result[key] = value_type(value).parse()

        return result

    def run(self):
        tokens = self.tokenize()
        result = self.parse(tokens)
        return result



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime
import uuid

import click

from caldavctl import dav

from icalendar import Todo


@click.command('list')
@click.option('-d', '--description', is_flag=True, show_default=True, default=False,
              help='Show the event\'s description.')
@click.option('-a', '--all', is_flag=True, show_default=True, default=False,
              help='Show all todos, including completed todos.')
@click.pass_obj
def list_todos(context, description, all):
    '''
    List todos from the default server and default calendar
    '''
    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()
    with dav.caldav_calendar(server, calendar_id) as calendar:
        # Check if the calendar supports events:
        if 'VTODO' not in calendar.get_supported_components():
            raise click.UsageError(f'This calendar "{calendar.name}" '
                                   'does not support todos.')
        # Get events in time range
        todos = calendar.todos(include_completed=all)
        for todo in todos:
            td = todo.icalendar_component
            summary = td.get('summary')
            status = '[✓]' if td.get('status', '') else '[ ]'
            percentage = td.get('percent-complete', 0)
            desc = td.get('description', '')
            uid = td.get('uid')

            click.echo(f'{status} {percentage:3}% {summary} - {uid}')
            if description and desc.strip():
                click.echo(desc)


@click.command('create')
@click.argument('summary')
@click.option('--description')
@click.option('--due-date')
@click.option('--priority')
@click.pass_obj
def create_todo(context, summary, description, due_date, priority):
    '''Create a todo on the server'''
    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()
    tz = context['config'].tz()

    with dav.caldav_calendar(server, calendar_id) as calendar:
        todo = Todo()

        todo.add('uid', str(uuid.uuid4()))
        todo.add('summary', summary)
        if description:
            todo.add('description', description)
        if due_date:
            try:
                dt = tz.localize(datetime.fromisoformat(due_date))
            except ValueError:
                raise click.UsageError(f'Invalid value for due date "{due_date}"')
            todo.add('due', dt)
        if priority:
            todo.add('priority', priority)
        todo.add('created', datetime.now(tz))
        todo.add('last-modified', datetime.now(tz))

        # Convert the todo to a string
        todo_string = todo.to_ical().decode('utf-8')

        try:
            calendar.save_event(todo_string)
            click.echo(f"todo created successfully: {summary}")
        except Exception as e:
            click.echo(f"Error creating todo: {e}")


@click.command('delete')
@click.argument('todo_id')
@click.pass_obj
def delete_todo(context, todo_id):
    '''Delete a todo on the server'''
    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()

    with dav.caldav_calendar_todo(server, calendar_id, todo_id) as todo:
        todo.delete()
    click.echo('Todo deleted')


@click.command('toggle')
@click.argument('todo_id')
@click.pass_obj
def toggle_todo_complete(context, todo_id):
    '''Toggle todo completed'''
    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()
    tz = context['config'].tz()

    with dav.caldav_calendar_todo(server, calendar_id, todo_id) as todo:
        if todo.icalendar_component.get('status', ''):
            del todo.icalendar_component['status']
            del todo.icalendar_component['completed']
        else:
            todo.icalendar_component.add('status', 'COMPLETED')
            todo.icalendar_component.add('completed', datetime.now(tz))
        todo.save()
        click.echo('Todo completed')


@click.command('percentage')
@click.argument('todo_id')
@click.argument('percent', type=click.IntRange(0, 100))
@click.pass_obj
def percentage_complete(context, todo_id, percent):
    '''Set percentage completed'''
    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()
    tz = context['config'].tz()

    with dav.caldav_calendar_todo(server, calendar_id, todo_id) as todo:
        if percent == todo.icalendar_component.get('percent-complete', 0):
            click.echo('No change!')
            return
        if 'percent-complete' in todo.icalendar_component:
            del todo.icalendar_component['percent-complete']
        todo.icalendar_component.add('percent-complete', percent)
        if todo.icalendar_component.get('status', ''):
            del todo.icalendar_component['status']
            del todo.icalendar_component['completed']
        if percent == 100:
            todo.icalendar_component.add('status', 'COMPLETED')
            todo.icalendar_component.add('completed', datetime.now(tz))
        todo.save()
        click.echo('Percentage set.')



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from click.testing import CliRunner
from unittest.mock import patch, MagicMock
from caldavctl.todo import (
    list_todos, create_todo, delete_todo, toggle_todo_complete, percentage_complete
)


def test_list_todos():
    runner = CliRunner()
    mock_calendar = MagicMock()
    mock_calendar.todos.return_value = [
        MagicMock(icalendar_component={
            'summary': 'Test Todo',
            'status': '',
            'percent-complete': 50,
            'description': 'Test Description',
            'uid': '1234'
        })
    ]

    with patch('caldavctl.dav.caldav_calendar', return_value=mock_calendar):
        result = runner.invoke(list_todos, ['--description'])

    assert result.exit_code == 0
    assert '[ ]  50% Test Todo - 1234' in result.output
    assert 'Test Description' in result.output


def test_create_todo():
    runner = CliRunner()
    mock_calendar = MagicMock()

    with patch('caldavctl.dav.caldav_calendar', return_value=mock_calendar):
        result = runner.invoke(
            create_todo, [
                'Test Todo',
                '--description', 'Test Description',
                '--due-date', '2024-12-31T23:59:59',
                '--priority', '1'
            ]
        )

    assert result.exit_code == 0
    assert 'todo created successfully: Test Todo' in result.output
    assert mock_calendar.save_event.called

# def test_delete_todo():
#     runner = CliRunner()
#     mock_todo = MagicMock()
#
#     with patch('caldavctl.dav.caldav_calendar_todo', return_value=mock_todo):
#         result = runner.invoke(delete_todo, ['1234'])
#
#     assert result.exit_code == 0
#     assert 'Todo deleted' in result.output
#     assert mock_todo.delete.called
#
# def test_toggle_todo_complete():
#     runner = CliRunner()
#     mock_todo = MagicMock()
#
#     # Test marking as completed
#     mock_todo.icalendar_component.get.return_value = ''
#
#     with patch('caldavctl.dav.caldav_calendar_todo', return_value=mock_todo):
#         result = runner.invoke(toggle_todo_complete, ['1234'])
#
#     assert result.exit_code == 0
#     assert 'Todo completed' in result.output
#     mock_todo.save.assert_called()
#
#     # Test marking as not completed
#     mock_todo.icalendar_component.get.return_value = 'COMPLETED'
#
#     with patch('caldavctl.dav.caldav_calendar_todo', return_value=mock_todo):
#         result = runner.invoke(toggle_todo_complete, ['1234'])
#
#     assert result.exit_code == 0
#     assert 'Todo completed' in result.output
#     mock_todo.save.assert_called()
#
# def test_percentage_complete():
#     runner = CliRunner()
#     mock_todo = MagicMock()
#     mock_todo.icalendar_component.get.side_effect = lambda key, default: {
#         'percent-complete': 50
#     }.get(key, default)
#
#     with patch('caldavctl.dav.caldav_calendar_todo', return_value=mock_todo):
#         result = runner.invoke(percentage_complete, ['1234', '75'])
#
#     assert result.exit_code == 0
#     assert 'Percentage set.' in result.output
#     mock_todo.save.assert_called()
